/**
* @overview A JavaScript prototype for creating world clocks.
* @requires jQuery
* @requires moment
* @requires moment-timezone
* @author Bart Busschots
* @license BSD-2-Clause
*/

//
// === Add needed JSDoc data type definitions ===
//

/**
* A TZ string from the [IANA Time Zone Database](https://en.wikipedia.org/wiki/Tz_database).
* A full list of the valid strings can be found in the
* [third column of this listing}(https://en.wikipedia.org/wiki/List_of_tz_database_time_zones).
* @typedef {string} TimeZoneString
*/

/**
* A jQuery object.
* @typedef {object} jQuery
*/

/**
* A jQuery object representing exactly one HTML span element.
* @typedef {jQuery} jQuerySingleSpan
*/

// make sure the needed pre-requisites are installed.
if(typeof jQuery !== 'function'){
    throw new Error('jQuery is required but not loaded');
}

// init the pbs namespace safely
/**
* APIs related to the [Programming by Stealth podcast/blog series](http://bartb.ie/pbs) are grouped under this namespace.
* @namespace
*/
var pbs = pbs ? pbs : {};

// add all the API's functionality within a self-executing anonymous function
(function(pbs, $, undefined){
    //
    // === Define private helper functions and their needed data structures ===
    //

    /**
    * A helper function for testing if a given value is a {@link jQuery}.
    * @memberof pbs
    * @inner
    * @access private
    * @param {*} $s - the value to test.
    * @returns {boolean} - `true` if the value being tested is a valid
    *     {@link jQuery}, `false` otherwise.
    */
    var isjQuery = function(obj){
        if(typeof obj !== 'object'){
            return false;
        }
        if(!(obj instanceof $)){
            return false;
        }
        return true;
    };

    /**
    * A helper function for testing if a given value is a {@link jQuerySingleSpan}
    * @memberof pbs
    * @inner
    * @access private
    * @param {*} $s - the value to test.
    * @returns {boolean} - `true` if the value being tested is a valid
    *     {@link jQuerySingleSpan}, `false` otherwise.
    */
    var isSingleSpan = function($s){
        if(!isjQuery($s)){
            return false;
        }
        if($s.length !== 1){
            return false;
        }
        return $s.is('span') ? true : false;
    };

    /**
    * A lookup table for validating TZ strings. All valid strings are keys in
    * this table with value `true`.
    * @inner
    * @private
    * @const
    * @memberof pbs
    * @type {Object.<string, boolean>}
    */
    var tzLookup = {};
    moment.tz.names().forEach(function(tzn){
        tzLookup[tzn] = true;
    });

    /**
    * A helper function for testing if a given value is valid
    * {@link TimeZoneString}.
    * @memberof pbs
    * @inner
    * @private
    * @param {*} tz - the value to test.
    * @returns {boolean} - `true` if the value being tested is a valid
    *     {@link TimeZoneString}, `false` otherwise.
    */
    var isValidTimeZone = function(tz){
        if(typeof tz !== 'string'){
            return false;
        }
        return tzLookup[tz] ? true : false;
    };

    //
    // === Define a prototype to represent a Single World Clock ===
    //

    /**
    * @constructor
    * @classdesc A prototype to represent a single world clock.
    *
    * The timezone can be specified either as an argument to the constructor, or
    * by specifying it with the data attribute `data-timezone` on the `span` to
    * be converted into a clock. An argument to the constructor takes precedence
    * over the data attribute.
    *
    * The markup produced to represent the clock is very simple:
    *
    * ```
    * <span class="pbs-worldclock">
    *  <span class="pbs-worldclock-hours">HH</span>
    *  <span class="pbs-worldclock-separator">:</span>
    *  <span class="pbs-worldclock-minutes">MM</span>
    * </span>
    * ```
    *
    * No CSS attributes are set on any of the elements, so all styling is left
    * to the user. Below is a sample style:
    *
    * ```
    * .pbs-worldclock{
    *    font-weight: bold;
    *    display: inline-block;
    *    border: 1px solid green;
    *    border-radius: 3px;
    *    padding: 3px;
    *    background-color: black;
    *    color: lightgreen;
    *    font-family: monospace;
    * }
    * ```
    *
    * @param {jQuerySingleSpan} $span - a jQuery object representing the HTML
    *     `span` element to be transformed into a clock. Note that all existing
    *     content within the span will be  will be removed.
    * @param {TimeZoneString} [tz=Europe/London] - The timezone for the clock.
    * @throws {TypeError} An error is thrown when an invalid value is passed for
    *     any of the arguments.
    */
    pbs.WorldClock = function($span, tz){
        // make sure we were passed a jQuery object representing exactly one span
        if(!isSingleSpan($span)){
            throw new TypeError('the first argument must be a jQuery object representing exactly one span element');
        }

        // make sure the span has not already been initialised as a clock
        if($span.is('.pbs-worldclock')){
            throw new Error('Cannot initialise a World Clock into a span that has already been initialised as a World Clock');
        }

        // save a reference to the span into the object
        /**
        * A jQuery object representing the span containing the clock.
        * @member {jQuerySingleSpan}
        * @private
        */
        this._$span = $span;

        // figure out which timezone to use
        /**
        * The clock's timezone as a string.
        * @member {TimeZoneString}
        * @private
        */
        this._timezone = 'Europe/London'; // default to Greenwich
        if(typeof tz === 'string'){
            // if there is a valid second argument, use it
            this._timezone = tz;
        }else if($span.data('timezone')){
            // if there is no second argument, but
            // there is a data attribute, use it
            this._timezone = $span.data('timezone');
        }

        // validate the timezone
        if(!isValidTimeZone(this._timezone)){
            throw new TypeError('Invalid timezone string: ' + this._timezone);
        }

        // initialise a placeholder for the interval ID
        /**
        * When the clock is running, the ID of the interval controlling it,
        * otherwise, 0.
        * @member {integer}
        * @private
        */
        this._intervalId = 0; // will hold the ID for the interval

        // initialise the span
        this._$span.empty().addClass('pbs-worldclock');
        /**
        * The inner span for the hours.
        * @member {jQuerySingleSpan}
        * @private
        */
        this._$hours = $('<span />').addClass('pbs-worldclock-hours');
        this._$span.append(this._$hours);
        /**
        * The inner span for the separator between the hours and minutes.
        * @member {jQuerySingleSpan}
        * @private
        */
        this._$separatorHM = $('<span />').text(':').addClass('pbs-worldclock-separator');
        this._$span.append(this._$separatorHM);
        /**
        * The inner span for the minutes.
        * @member {jQuerySingleSpan}
        * @private
        */
        this._$minutes = $('<span />').addClass('pbs-worldclock-minutes');
        this._$span.append(this._$minutes);

        // save a reference to this object into the span
        this._$span.data('pbsWorldclock', this);

        // start the clock
        this.start();
    };

    //
    // -- Accessor Methods --
    //

    /**
    * Get and/or set the clock's timezone.
    * @param {TimeZoneString} [tz]
    * @returns {TimeZoneString}
    * @throws {TypeError} An error is thrown if an argument is passed that is
    *     not valid.
    */
    pbs.WorldClock.prototype.timezone = function(){
        // if there is a first argument, try use it as a timezone
        if(arguments.length >= 1){
            if(!isValidTimeZone(arguments[0])){
                throw new TypeError('invalid timezone');
            }
            this._timezone = arguments[0];
        }

        // always return the current timezone
        return this._timezone;
    };

    //
    // -- Methods for Rendering the Clock --
    //

    /**
    * Render the current time into a given clock.
    * @memberof pbs
    * @inner
    * @private
    * @param {pbs.WorldClock} clock - a reference to the clock to render the
    *     time into.
    */
    var renderClock = function(clock){
        // get the current time
        var now = moment().tz(clock._timezone);

        // render the current time
        clock._$hours.text(now.format('HH'));
        clock._$minutes.text(now.format('mm'));

        // blink the separator
        if(parseInt(now.format('ss')) %2 == 0){
            $('span.pbs-worldclock-separator', clock._$span).fadeTo(500, 0.2);
        }else{
            $('span.pbs-worldclock-separator', clock._$span).fadeTo(500, 1);
        }
    };

    /**
    * Start the clock.
    */
    pbs.WorldClock.prototype.start = function(){
        // if the clock is already started, do nothing
        if(this._intervalId !== 0){
            return;
        }

        // render the current time
        renderClock(this);

        // start an interval
        var self = this;
        this._intervalId = setInterval(function(){ renderClock(self); }, 1000);
    };

    /**
    * Stop the clock.
    */
    pbs.WorldClock.prototype.stop = function(){
        // if the clock is already stopped, do nothing
        if(this._intervalId === 0){
            return;
        }

		// stop the clock
        clearInterval(this._intervalId);
        this._intervalId = 0;
    };

    //
    // === Provide Automation ===
    //

    /**
    * Initialise all spans with the class `pbs-worldclock-auto` within a given
    * set of containers.
    * @param {jQueryObject} [$containers=$(document)] - the container(s) to
    *     search for spans to be automatially transformed into clocks. By
    *     default the entire document is searched.
    * @throws {TypeError} An error is thrown if the first argument is present,
    *     but not a jQuery object.
    */
    pbs.WorldClock.autoInitialise = function($containers){
        // default the container if none was passed
        if(typeof $containers === 'undefined'){
            $containers = $(document);
        }

        // make sure we have a jQuery object to search within
        if(!isjQuery($containers)){
            throw new TypeError('the first argument must be a jQuery object');
        }

        // search the container(s) and initialise each clock found
        $('span.pbs-worldclock-auto', $containers).each(function(){
            var $span = $(this);

            // initialise the clock
            new pbs.WorldClock($span);

            // remove the auto class to avoid re-initialisation
            $span.removeClass('pbs-worldclock-auto');
        });
    };

    // add an event handler to automaticlaly initialise clocks when the document
    // becomes ready
    $(function(){ pbs.WorldClock.autoInitialise(); });
})(pbs, jQuery);
